<?php
namespace Ep\Gath;

use Ep\Gath\Plugin\Products;
use Ep\Http\Middleware\Cache;
use GuzzleHttp\Client;
use GuzzleHttp\ClientInterface;
use RuntimeException;
use Throwable;
use Laminas\Log\LoggerInterface;
use Laminas\Stdlib\ArrayUtils;

/**
 * Class Gath
 *
 * @package Ep\Gath
 * @author Elektronika Perhoč <support@ep.hr>
 * @link http://gitlab.ep.local/ep-framework/gath
 */
class Gath
{
    const SORT_DATE_DESC  = 1;
    const SORT_PRICE_ASC  = 2;
    const SORT_PRICE_DESC = 3;
    const SORT_STOCK_DESC = 4;
    const SORT_NAME_DESC  = 5;
    const SORT_NAME_ASC   = 6;

    const ID_OFFSET = 1e7;
    const ID_OFFSET_SIMPLE = self::ID_OFFSET;
    const ID_OFFSET_GROUPED = self::ID_OFFSET_SIMPLE + 1e7;
    const ID_OFFSET_VARIABLE = self::ID_OFFSET_GROUPED + 1e7;
    const ID_OFFSET_VARIATION = self::ID_OFFSET_VARIABLE + 1e7;
    const ID_OFFSET_TIRES = self::ID_OFFSET_VARIATION + 1e7;

    /** @var array */
    protected $config;
    /** @var ClientInterface|Client */
    protected $http;
    /** @var LoggerInterface */
    protected $logger;
    private $attributes = [];

    private int $requestCount = 0;

    /**
     * Gath constructor.
     *
     * @param array $config
     * @param ClientInterface $http
     * @param LoggerInterface $logger
     */
    public function __construct(array $config, ClientInterface $http, LoggerInterface $logger)
    {
        $this->config = $config;
        $this->http   = $http;
        $this->logger = $logger;
    }

    public function getConfig(): array
    {
        return $this->config;
    }

    /**
     * Da li smo u debug modu?
     *
     * @return bool
     */
    public function isDebug(): bool
    {
        return $this->config['debug'] ?? false;
    }

    /**
     * Da li je ovo veleprodajni shop?
     *
     * @return bool
     */
    public function isWholesale(): bool
    {
        return $this->config['wholesale'] ?? false;
    }

    public function areDescriptionsFiltered(): bool
    {
        return $this->config['filter_descriptions'] ?? true;
    }

    public function usesWooCategories(): bool
    {
        return (bool) $this->config['woo_categories'];
    }

    public function getAttributes($categoryId = null): array
    {
        $key = md5(serialize($categoryId));

        if (!isset($this->attributes[$key])) {
            if ($categoryId) {
                $result = $this->request('POST', 'category/attributes', ['Categories' => $categoryId]);
            } else {
                $result = $this->request('GET', 'product/attributes');
            }

            $this->attributes[$key] = array_map(function (array $attribute) {
                $attribute['slug'] = sanitize_title($attribute['naziv']);
                return $attribute;
            }, $result['Results']);
        }

        return $this->attributes[$key];
    }

    /**
     * Vraća osnovni uri servisa sa slikama.
     *
     * @return string
     */
    public function getImageBaseUri(): string
    {
        return $this->config['image_base_uri'] ?? '';
    }

    /**
     * Da li prikazujemo kratki opis.
     *
     * @return bool
     */
    public function getShowShortDescription(): bool
    {
        return $this->config['show_short_description'] ?? true;
    }

    /**
     * @param int $categoryId
     *
     * @return array
     */
    public function getBrandsByCategory(int $categoryId): array
    {
        if ($categoryId > Products::TERM_OFFSET_CATEGORY) {
            $categoryId = intval($categoryId - Products::TERM_OFFSET_CATEGORY);
        }

        return $this->request('POST', 'product/brands', [
            'CategoryDataLink' => $categoryId,
        ]);
    }

    public function getBrands():  array
    {
        return $this->request('GET', 'product/brands');
    }

    public function getCategoryById($id): array
    {
        $params['DataLink'] = $id;
        return $this->request('GET', 'category', $params);
    }

    public function getCategoryBySlug($slug): array
    {
        $slug = is_array($slug) ? $slug[0] : $slug;
        $params = ['slug' => $slug];
        return $this->request('GET', 'category', $params);
    }
    
    public function getCategories($parent, int $depth): array
    {
        $params = ['Depth' => $depth];

        if ($parent) {
            $params['Parent'] = $parent;
        }

        return $this->request('POST', 'category', $params);
    }

    public function getAllCategories(): array
    {
        $results = $this->request('POST', 'category', ['Depth' => 1000]);
        return $results['Results'] ?? [];
//        return $this->flattenChildren($results['Results']);
    }

    private function flattenChildren(array $results): array
    {
        $flattened = [];

        foreach ($results as &$result) {
            if (isset($result['Children'])) {
                $children = $this->flattenChildren($result['Children'] ?? []);
                unset($result['Children']);
                $flattened[] = $result;
                $flattened = ArrayUtils::merge($flattened, $children);
            } else {
                $flattened[] = $result;
            }
        }

        return $flattened;
    }
    
    /**
     * Vraća istaknute proizvode.
     *
     * @param int $orderBy
     * @param int $limit
     * @param int $offset
     *
     * @return array
     */
    public function getFeaturedProducts(int $orderBy = 1, int $limit = 10, int $offset = 0): array
    {
        return $this->request('POST', 'products/featured', [
            'OrderBy' => $orderBy,
            'Limit' => $limit,
            'Offset' => $offset,
        ]);
    }

    /**
     * Vraća proizvođače motocikala.
     *
     * @return array
     */
    public function getMotorcycleBrands(): array
    {
        return $this->request('GET', 'moto/brands');
    }

    /**
     * Vraća modele motocikala za datog proizvođača.
     *
     * @param int $motorcycleBrandId
     *
     * @return array
     */
    public function getMotorcycleModels(int $motorcycleBrandId): array
    {
        return $this->request('POST', 'moto/models', [
            'Brands' => [$motorcycleBrandId],
        ]);
    }

    /**
     * @param int $motorcycleId
     *
     * @return array|null
     */
    public function getMotorcycle(int $motorcycleId): ?array
    {
        $motorcycles = $this->request('GET', 'moto/motorcycles', [
            'Id' => $motorcycleId,
        ]);

        return count($motorcycles['Results']) > 0 ? $motorcycles['Results'][0] : null;
    }

    /**
     * Vraća motocikle za dati model.
     *
     * @param int $modelId
     *
     * @return array
     */
    public function getMotorcyclesByModel(int $modelId): array
    {
        return $this->request('POST', 'moto/motorcycles', [
            'Models' => [$modelId],
        ]);
    }

    /**
     * Vraća cijene po šiframa proizvoda.
     *
     * @param array $skus
     * @param array $tireIds
     * @return array
     */
    public function getPricesBySkus(array $skus, array $tireIds = []): array
    {
        $user = wp_get_current_user();
        $username = $user->ID > 0 ? $user->user_login : null;

        $params = [
            'Username' => $username,
            'PriceRabatePolicy' => $this->config['wholesale'] ?? false,
            'Discount' => true,
            'Codes' => $skus,
        ];

        if ($this->config['use_price30']) {
            $params['LowestPriceInDays'] = 30;
        }

        if (!empty($tireIds)) {
            $params['TireIds'] = $tireIds;
        }

        $result = $this->request('POST', 'prices/by-codes', $params);

        if ($this->config['convert_prices_to_euro'] && ($exchange = $this->config['euro_exchange'])) {
            $result['Results'] = array_map(
                fn (array $price) => array_merge($price, [
                    'Currency' => 'EUR',
                    'PriceRetail' => round($price['PriceRetail'] / $exchange, 2),
                    'PriceWholesale' => round($price['PriceWholesale'] / $exchange, 2),
                    'PriceRetailMasa' => round($price['PriceRetailMasa'] / $exchange, 2),
                    'RabPolicy' => array_map(
                        fn (array $rebate) => array_merge($rebate, [
                            'PriceWholesale' => round($rebate['PriceWholesale'] / $exchange, 2),
                            'PriceWholesaleNetto' => round($rebate['PriceWholesaleNetto'] / $exchange, 2),
                            'CalculatedNetto' => round($rebate['CalculatedNetto'] / $exchange, 2),
                        ]),
                        $price['RabPolicy'] ?? [],
                    ),
                    'Discount' => array_map(
                        fn (array $discount) => array_merge($discount, [
                            'OriginalRetail' => round($discount['OriginalRetail'] / $exchange, 2),
                            'NewRetail' => round($discount['NewRetail'] / $exchange, 2),
                            'OriginalWholesale' => round($discount['OriginalWholesale'] / $exchange, 2),
                            'NewWholesale' => round($discount['NewWholesale'] / $exchange, 2),
                        ]),
                        $price['Discount'] ?? [],
                    ),
                ]),
                $result['Results'],
            );
        }

        return $result;
    }

    public function getProductsByAttributes(
        int $categoryId,
        array $attributes,
        array $brands,
        int $sort = self::SORT_DATE_DESC,
        int $limit = 10,
        int $offset = 0
    ): array {

        $aParams = [];
        foreach ($attributes as $id => $values) {
            $aParams[] = [
                'id' => $id,
                'values' => $values,
            ];
        }

        $params = [
            'Categories'      => [$categoryId],
            'Attributes'      => $aParams,
            'OrderBy'         => $sort,
            'Limit'           => $limit,
            'Offset'          => $offset,
            'SortNoStockLast' => $this->config['sort_no_stock_last'] ?? false,
        ];

        if (!empty($brands)) {
            $params['BrandIds'] = $brands;
        }

        return $this->request('POST', 'product/byattributes', $params);
    }

    /**
     * Vraća proizvode po wooIdu kategorije.
     *
     * @param int $categoryId
     * @param int $sort
     * @param int $limit
     * @param int $offset
     *
     * @param array|null $brands
     * @return array
     */
    public function getProductsByCategory(
        int $categoryId,
        int $sort = self::SORT_DATE_DESC,
        int $limit = 10,
        int $offset = 0,
        array $brands = null
    ): array {
        $params = [
            'OrderBy'         => $sort,
            'Limit'           => $limit,
            'Offset'          => $offset,
            'SortNoStockLast' => $this->config['sort_no_stock_last'] ?? false,
        ];

        if ($this->usesWooCategories()) {
            $params['CategoryId'] = $categoryId;
        } else {
            $params['CategoryDataLink'] = $categoryId;
        }

        if (null !== $brands && !empty($brands)) {
            $params['BrandIds'] = array_map('intval', array_values($brands));
        }

        return $this->request('POST', 'product/bycategory', $params);
    }

    public function getTires(
        array $filters,
        int $order = self::SORT_DATE_DESC, int $limit = 10, int $offset = 0
    ): array
    {
        $params = [
            'Limit' => $limit,
            'Offset' => $offset,
            'OrderBy' => $order,
            'SortNoStockLast' => $this->config['sort_no_stock_last'] ?? false,
        ];

        if (!empty($filters['ArtApply'])) $params['ArtApply'] = [$filters['ArtApply']];
        if (!empty($filters['TireWidth'])) $params['TireWidth'] = (int)$filters['TireWidth'];
        if (!empty($filters['TireHeight'])) $params['TireHeight'] = (int)$filters['TireHeight'];
        if (!empty($filters['TireDiameter'])) $params['TireDiameter'] = (float)$filters['TireDiameter'];
        if (!empty($filters['Season'])) $params['Season'] = [ $filters['Season'] ];
        if (!empty($filters['VehicleType'])) $params['VehicleType'] = [ $filters['VehicleType'] ];
        if (!empty($filters['BrandName'])) $params['BrandName'] = [ $filters['BrandName'] ];
        if (!empty($filters['LoadIndex1'])) $params['LoadIndex1'] = [ (int)$filters['LoadIndex1'] ];
        if (!empty($filters['SpeedIndex'])) $params['SpeedIndex'] = [ $filters['SpeedIndex'] ];

//        var_dump($params);
        return $this->request('POST', 'product/tyres', $params);
    }

    /**
     * Vraća proizvode po id-u.
     *
     * @param array $productIds
     *
     * @param int $limit
     * @param int $offset
     * @return array
     */
    public function getProductsByIds(array $productIds, int $limit = 1000, int $offset = 0): array
    {
        $dataLinks = [
            'TireIds'     => [],
            'SimpleIds'   => [],
            'GroupedIds'  => [],
            'VariableIds' => [],
            'VariationIds' => [],
            'Limit'       => $limit,
            'Offset'      => $offset,
        ];

        // pretvaramo id-eve u DataLinkove i razdvajamo po vrsti proizvoda.
        foreach ($productIds as $productId) {
            if ($productId > static::ID_OFFSET_TIRES) {
                $dataLinks['TireIds'][] = intval($productId - static::ID_OFFSET_TIRES);
            } else if ($productId > static::ID_OFFSET_VARIATION) {
                $dataLinks['VariationIds'][] = intval($productId - static::ID_OFFSET_VARIATION);
            } else if ($productId > static::ID_OFFSET_VARIABLE) {
                $dataLinks['VariableIds'][] = intval($productId - static::ID_OFFSET_VARIABLE);
            } else if ($productId > static::ID_OFFSET_GROUPED) {
                $dataLinks['GroupedIds'][] = intval($productId - static::ID_OFFSET_GROUPED);
            } else {
                $dataLinks['SimpleIds'][] = intval($productId - static::ID_OFFSET_SIMPLE);
            }
        }

        $result = [];

        foreach (['TireIds', 'GroupedIds', 'SimpleIds', 'VariationIds', 'VariableIds'] as $type) {
            if (count($dataLinks[$type]) === 0) {
                continue;
            }

            $result = ArrayUtils::merge($result, $this->request('POST', 'product/byids', [
                $type => $dataLinks[$type],
                'Limit' => $dataLinks['Limit'],
                'Offset' => $dataLinks['Offset'],
            ]));
        }

        // Ovo ne valja nikako
        $charConvert = [
            'Ĺľ' => 'ž',
            '&scaron;' => 'š',
            'ÄŤ' => 'č',
            'Ä‡' => 'ć',
            'Ä‘' => 'đ',
            'Ä†' => 'Č',
            'ÄŚ' => 'Ć',
            'Ĺ˝' => 'Ž'
        ];
        foreach ($result['Results'] as &$res) {
            if (isset($res['description'])) {
                $res['description'] = str_replace(array_keys($charConvert), array_values($charConvert), $res['description']);
//                $res['description'] = strip_tags($res['description'], '<span>');

            }

        }

        return $result;
    }

    /**
     * Vraća simple proizvode po šiframa.
     *
     * @param array $skus
     *
     * @return array
     */
    public function getProductsByskus(array $skus): array
    {
        return $this->request('POST', 'product/byids', [
            'SimpleCodes' => $skus,
            'Limit' => 1000,
        ]);
    }

    /**
     * @param array $tecDocPairs
     * @param int $order
     * @param int $limit
     * @param int $offset
     *
     * @return mixed
     */
    public function getProductsByTecDocPairs(array $tecDocPairs, int $order, int $limit = 10, int $offset = 0): array
    {
        $params = [
            'TecDocKeys'      => $tecDocPairs,
            'OrderBy'         => $order,
            'Limit'           => $limit,
            'Offset'          => $offset,
            'SortNoStockLast' => $this->config['sort_no_stock_last'] ?? false,
        ];

        return $this->request('POST', 'product/bytecdocids', $params);
    }

    /**
     * @param $motorcycleId
     * @param $genericArticleIds
     * @param int $orderBy
     * @param int $limit
     * @param int $offset
     *
     * @return array
     */
    public function getProductsByMotorcycle(
        int $motorcycleId,
        array $genericArticleIds,
        int $orderBy,
        int $limit = 10,
        int $offset = 0
    ): array {
        return $this->request('POST', 'moto/articles', [
            'MotoId' => $motorcycleId,
            'GenArtIds' => $genericArticleIds,
            'OrderBy' => $orderBy,
            'Limit' => $limit,
            'Offset' => $offset,
            'SortNoStockLast' => $this->config['sort_no_stock_last'] ?? false,
        ]);
    }

    /**
     * Vraća stanje zalihe po šifri proizvoda.
     *
     * @param array $skus
     *
     * @param array $tireIds
     * @return array
     */
    public function getStockBySkus(array $skus, array $tireIds = []): array
    {
        $params = [
            'Codes' => $skus,
            'StockByWarehouses' => $this->config['stock_by_warehouses'] ?? false,
        ];

        if (!empty($tireIds)) {
            $params['TireIds'] = $tireIds;
        }

        return $this->request('POST', 'stock/by-codes', $params);
    }

    public function getWishlistProducts(string $username): array
    {
        return $this->request('POST', 'favoriteproducts', [
            'Function' => 'Read',
            'Username' => $username,
        ]);
    }

    /**
     * Upućuje zahtjev prema servisu.
     *
     * @param $method
     * @param $path
     * @param array $params
     *
     * @return array
     */
    public function request(string $method = 'GET', string $path = '', array $params = []): array
    {
        $session_id = session_id() . ' ';

        try {
            if ('POST' === $method) {
                $request = [
                    'json' => $params,
                    'headers' => [
                        'Content-Type' => 'application/json',
                        'Connection' => 'keep-alive',
                        'Keep-Alive' => 'timeout=5, max=100',
                    ],
                ];
            } else {
                $request = [
                    'query' => $params,
                    'headers' => [
                        'Connection' => 'keep-alive',
                        'Keep-Alive' => 'timeout=5, max=100',
                    ],
                ];
            }

            try {
                $this->requestCount++;
                do_action('gath_request_pre', $this->requestCount, $this->config['base_uri'] . $path, $params, $this);
                if ($this->isDebug()) {
                    ob_start();
                    $response = $this->http->request($method, $path, $request);
                    $this->logger->info($session_id . 'CACHE :: ' . $response->getHeaderLine(Cache::HEADER_CACHE_INFO));
                    $debug    = ob_get_clean();
                    $this->logger->debug($debug);
                } else {
                    $response = $this->http->request($method, $path, $request);
                }
            } catch (Throwable $e) {
                do_action('gath_error', $this->requestCount, $e->getMessage(), $this);
                $this->logger->crit($session_id . $e->getMessage() . "\n" . json_encode($params));
                throw new RuntimeException($e->getMessage());
            }

            if (300 <= $response->getStatusCode()) {
                do_action('gath_error', $this->requestCount, $response->getReasonPhrase(), $this);
                $this->logger->err($session_id . sprintf('Došlo je do greške s Gathom %s: %s %s', $response->getReasonPhrase(), $path, json_encode($params)));
                throw new RuntimeException('Gath greška: ' . $response->getReasonPhrase());
            }

            $jsonContent = $response->getBody()->getContents();
            $data = json_decode($jsonContent, true);

            if (null === $data) {
                $this->logger->err($session_id . sprintf(
                    "Greška prilikom dekodiranja JSON-a: %s.\nJSON: %s",
                    json_last_error_msg(),
                    $jsonContent
                ));
                throw new RuntimeException('Greška prilikom dekodiranja JSON-a.');
            }

            if (!isset($data['Results']) && !isset($data['Result'])) {
                throw new RuntimeException("Objekt iz Gatha ne sadrži polje 'Results'.");
            }

            if (!is_array($data['Results'] ?? []) && !is_array($data['Result'] ?? [])) {
                throw new RuntimeException(sprintf(
                    "Polje 'Results' mora sadržavati array, a ne %s",
                    is_object($data) ? get_class($data['Results']) : gettype($data['Results'])
                ));
            }

            do_action('gath_request_post', $this->requestCount, $data, $this);
        } catch (Throwable $e) {
            if (!preg_match('/Nemam kodova/', $e->getMessage()) && $this->config['throw_exceptions']) {
                die($e->getMessage());
            }

            $data = [
                'Results' => [],
                'RecordCount' => 0,
            ];
        }
        
        return $data;
    }
}
